import numpy as np
from astropy import cosmology as cosmo

import autofit as af
from autolens import exc
from autolens.lens import ray_tracing, lens_data as ld, lens_fit
from autolens.model.galaxy import galaxy as g
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.pipeline import phase_tagging
from autolens.pipeline.phase import phase_extensions
from autolens.pipeline.phase.phase import Phase, setup_phase_mask
from autolens.pipeline.plotters import phase_plotters


def isinstance_or_prior(obj, cls):
    if isinstance(obj, cls):
        return True
    if isinstance(obj, af.PriorModel) and obj.cls == cls:
        return True
    return False


class PhaseImaging(Phase):
    hyper_image_sky = af.PhaseProperty("hyper_image_sky")
    hyper_background_noise = af.PhaseProperty("hyper_background_noise")
    galaxies = af.PhaseProperty("galaxies")

    def __init__(
        self,
        phase_name,
        phase_folders=tuple(),
        galaxies=None,
        hyper_image_sky=None,
        hyper_background_noise=None,
        optimizer_class=af.MultiNest,
        sub_grid_size=2,
        signal_to_noise_limit=None,
        bin_up_factor=None,
        image_psf_shape=None,
        inversion_psf_shape=None,
        positions_threshold=None,
        mask_function=None,
        inner_mask_radii=None,
        interp_pixel_scale=None,
        use_inversion_border=True,
        inversion_pixel_limit=None,
        cluster_pixel_scale=None,
        cosmology=cosmo.Planck15,
        auto_link_priors=False,
    ):

        """

        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit models and hyper_galaxy
        passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        cluster_pixel_scale : float or None
            If *True*, the hyper_galaxy image used to generate the cluster'grids weight map will be binned up to this \
            higher pixel scale to speed up the KMeans clustering algorithm.
        """

        phase_tag = phase_tagging.phase_tag_from_phase_settings(
            sub_grid_size=sub_grid_size,
            signal_to_noise_limit=signal_to_noise_limit,
            bin_up_factor=bin_up_factor,
            image_psf_shape=image_psf_shape,
            inversion_psf_shape=inversion_psf_shape,
            positions_threshold=positions_threshold,
            inner_mask_radii=inner_mask_radii,
            interp_pixel_scale=interp_pixel_scale,
            cluster_pixel_scale=cluster_pixel_scale,
        )

        super(PhaseImaging, self).__init__(
            phase_name=phase_name,
            phase_tag=phase_tag,
            phase_folders=phase_folders,
            optimizer_class=optimizer_class,
            cosmology=cosmology,
            auto_link_priors=auto_link_priors,
        )

        self.sub_grid_size = sub_grid_size
        self.signal_to_noise_limit = signal_to_noise_limit
        self.bin_up_factor = bin_up_factor
        self.image_psf_shape = image_psf_shape
        self.inversion_psf_shape = inversion_psf_shape
        self.positions_threshold = positions_threshold
        self.mask_function = mask_function
        self.inner_mask_radii = inner_mask_radii
        self.interp_pixel_scale = interp_pixel_scale
        self.use_inversion_border = use_inversion_border
        self.inversion_pixel_limit = inversion_pixel_limit
        self.cluster_pixel_scale = cluster_pixel_scale

        inversion_pixel_limit_from_prior = int(
            af.conf.instance.prior_default.get(
                "pixelizations", "VoronoiBrightnessImage", "pixels"
            )[2]
        )

        if self.inversion_pixel_limit is not None:

            self.cluster_pixel_limit = min(
                inversion_pixel_limit_from_prior, self.inversion_pixel_limit
            )

        else:

            self.cluster_pixel_limit = inversion_pixel_limit_from_prior
            self.inversion_pixel_limit = self.cluster_pixel_limit

        self.galaxies = galaxies or []
        self.hyper_image_sky = hyper_image_sky
        self.hyper_background_noise = hyper_background_noise

        self.hyper_noise_map_max = af.conf.instance.general.get(
            "hyper", "hyper_noise_map_max", float
        )

    @property
    def uses_hyper_images(self):
        if self.galaxies:
            for galaxy in self.galaxies:
                if galaxy.hyper_galaxy is not None or isinstance_or_prior(
                    galaxy.regularization, reg.AdaptiveBrightness
                ):
                    return True
        return False

    @property
    def uses_inversion(self):
        if self.galaxies:
            for galaxy in self.galaxies:
                if galaxy.pixelization is not None:
                    return True
        return False

    @property
    def uses_cluster_inversion(self):
        if self.galaxies:

            for galaxy in self.galaxies:
                if isinstance_or_prior(galaxy.pixelization, pix.VoronoiBrightnessImage):
                    return True

        return False

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def modify_image(self, image, results):
        """
        Customize an lens_data. e.g. removing lens light.

        Parameters
        ----------
        image: scaled_array.ScaledSquarePixelArray
            An lens_data that has been masked
        results: autofit.tools.pipeline.ResultsCollection
            The result of the previous lens

        Returns
        -------
        lens_data: scaled_array.ScaledSquarePixelArray
            The modified image (not changed by default)
        """
        return image

    def run(self, data, results=None, mask=None, positions=None):
        """
        Run this phase.

        Parameters
        ----------
        positions
        mask: Mask
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            An object describing the results of the last phase or None if no phase has been executed
        data: scaled_array.ScaledSquarePixelArray
            An lens_data that has been masked

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper_galaxy.
        """
        analysis = self.make_analysis(
            data=data, results=results, mask=mask, positions=positions
        )

        self.pass_priors(results)
        self.assert_and_save_pickle()

        result = self.run_analysis(analysis)

        return self.make_result(result, analysis)

    def make_analysis(self, data, results=None, mask=None, positions=None):
        """
        Create an lens object. Also calls the prior passing and lens_data modifying functions to allow child
        classes to change the behaviour of the phase.

        Parameters
        ----------
        positions
        mask: Mask
            The default masks passed in by the pipeline
        data: im.CCD
            An lens_data that has been masked
        results: autofit.tools.pipeline.ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens : Analysis
            An lens object that the non-linear optimizer calls to determine the fit of a set of values
        """

        mask = setup_phase_mask(
            data=data,
            mask=mask,
            mask_function=self.mask_function,
            inner_mask_radii=self.inner_mask_radii,
        )

        if self.positions_threshold is not None and positions is not None:
            positions = list(
                map(lambda position_set: np.asarray(position_set), positions)
            )
        elif self.positions_threshold is None:
            positions = None
        elif self.positions_threshold is not None and positions is None:
            raise exc.PhaseException(
                "You have specified for a phase to use positions, but not input positions to the "
                "pipeline when you ran it."
            )

        lens_data = ld.LensData(
            ccd_data=data,
            mask=mask,
            sub_grid_size=self.sub_grid_size,
            image_psf_shape=self.image_psf_shape,
            positions=positions,
            interp_pixel_scale=self.interp_pixel_scale,
            cluster_pixel_scale=self.cluster_pixel_scale,
            cluster_pixel_limit=self.cluster_pixel_limit,
            uses_inversion=self.uses_inversion,
            uses_cluster_inversion=self.uses_cluster_inversion,
            hyper_noise_map_max=self.hyper_noise_map_max,
        )

        modified_image = self.modify_image(
            image=lens_data.unmasked_image, results=results
        )

        lens_data = lens_data.new_lens_data_with_modified_image(
            modified_image=modified_image
        )

        if self.signal_to_noise_limit is not None:
            lens_data = lens_data.new_lens_data_with_signal_to_noise_limit(
                signal_to_noise_limit=self.signal_to_noise_limit
            )

        if self.bin_up_factor is not None:
            lens_data = lens_data.new_lens_data_with_binned_up_ccd_data_and_mask(
                bin_up_factor=self.bin_up_factor
            )

        self.output_phase_info()

        analysis = self.Analysis(
            lens_data=lens_data,
            cosmology=self.cosmology,
            positions_threshold=self.positions_threshold,
            image_path=self.optimizer.image_path,
            results=results,
            use_inversion_border=self.use_inversion_border,
            inversion_pixel_limit=self.inversion_pixel_limit,
        )

        return analysis

    def output_phase_info(self):

        file_phase_info = "{}/{}".format(self.optimizer.phase_output_path, "phase.info")

        with open(file_phase_info, "w") as phase_info:
            phase_info.write("Optimizer = {} \n".format(type(self.optimizer).__name__))
            phase_info.write("Sub-grid size = {} \n".format(self.sub_grid_size))
            phase_info.write("Image PSF shape = {} \n".format(self.image_psf_shape))
            phase_info.write(
                "Pixelization PSF shape = {} \n".format(self.inversion_psf_shape)
            )
            phase_info.write(
                "Positions Threshold = {} \n".format(self.positions_threshold)
            )
            phase_info.write("Cosmology = {} \n".format(self.cosmology))
            phase_info.write("Auto Link Priors = {} \n".format(self.auto_link_priors))

            phase_info.close()

    def extend_with_inversion_phase(self):
        return phase_extensions.InversionPhase(phase=self)

    def extend_with_multiple_hyper_phases(
        self,
        hyper_galaxy=False,
        inversion=False,
        include_background_sky=False,
        include_background_noise=False,
    ):

        if hyper_galaxy:
            if not include_background_sky and not include_background_noise:
                phase_hyper_galaxy = (
                    phase_extensions.hyper_galaxy_phase.HyperGalaxyPhase
                )
            elif include_background_sky and not include_background_noise:
                phase_hyper_galaxy = (
                    phase_extensions.hyper_galaxy_phase.HyperGalaxyBackgroundSkyPhase
                )
            elif not include_background_sky and include_background_noise:
                phase_hyper_galaxy = (
                    phase_extensions.hyper_galaxy_phase.HyperGalaxyBackgroundNoisePhase
                )
            else:
                phase_hyper_galaxy = (
                    phase_extensions.hyper_galaxy_phase.HyperGalaxyBackgroundBothPhase
                )
        else:
            phase_hyper_galaxy = None

        if inversion:
            if not include_background_sky and not include_background_noise:
                phase_inversion = phase_extensions.InversionPhase
            elif include_background_sky and not include_background_noise:
                phase_inversion = phase_extensions.InversionBackgroundSkyPhase
            elif not include_background_sky and include_background_noise:
                phase_inversion = phase_extensions.InversionBackgroundNoisePhase
            else:
                phase_inversion = phase_extensions.InversionBackgroundBothPhase
        else:
            phase_inversion = None

        hyper_phase_classes = tuple(filter(None, (phase_hyper_galaxy, phase_inversion)))

        if len(hyper_phase_classes) == 0:
            return self
        else:
            return phase_extensions.CombinedHyperPhase(
                phase=self, hyper_phase_classes=hyper_phase_classes
            )

    # noinspection PyAbstractClass
    class Analysis(Phase.Analysis):
        def __init__(
            self,
            lens_data,
            cosmology,
            positions_threshold,
            use_inversion_border=True,
            inversion_pixel_limit=None,
            image_path=None,
            results=None,
        ):

            super(PhaseImaging.Analysis, self).__init__(
                cosmology=cosmology, results=results
            )

            self.lens_data = lens_data

            self.positions_threshold = positions_threshold

            self.should_plot_image_plane_pix = af.conf.instance.visualize.get(
                "figures", "plot_image_plane_adaptive_pixelization_grid", bool
            )

            self.plot_data_as_subplot = af.conf.instance.visualize.get(
                "plots", "plot_data_as_subplot", bool
            )

            self.plot_data_image = af.conf.instance.visualize.get(
                "plots", "plot_data_image", bool
            )

            self.plot_data_noise_map = af.conf.instance.visualize.get(
                "plots", "plot_data_noise_map", bool
            )

            self.plot_data_psf = af.conf.instance.visualize.get(
                "plots", "plot_data_psf", bool
            )

            self.plot_data_signal_to_noise_map = af.conf.instance.visualize.get(
                "plots", "plot_data_signal_to_noise_map", bool
            )

            self.plot_data_absolute_signal_to_noise_map = af.conf.instance.visualize.get(
                "plots", "plot_data_absolute_signal_to_noise_map", bool
            )

            self.plot_data_potential_chi_squared_map = af.conf.instance.visualize.get(
                "plots", "plot_data_potential_chi_squared_map", bool
            )

            self.plot_lens_fit_all_at_end_png = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_all_at_end_png", bool
            )
            self.plot_lens_fit_all_at_end_fits = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_all_at_end_fits", bool
            )

            self.plot_lens_fit_as_subplot = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_as_subplot", bool
            )

            self.plot_lens_fit_of_planes_as_subplot = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_of_planes_as_subplot", bool
            )

            self.plot_lens_fit_inversion_as_subplot = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_inversion_as_subplot", bool
            )

            self.plot_lens_fit_image = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_image", bool
            )

            self.plot_lens_fit_noise_map = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_noise_map", bool
            )

            self.plot_lens_fit_signal_to_noise_map = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_signal_to_noise_map", bool
            )

            self.plot_lens_fit_model_image = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_model_image", bool
            )

            self.plot_lens_fit_residual_map = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_residual_map", bool
            )

            self.plot_lens_fit_normalized_residual_map = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_normalized_residual_map", bool
            )

            self.plot_lens_fit_chi_squared_map = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_chi_squared_map", bool
            )

            self.plot_lens_fit_contribution_maps = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_contribution_maps", bool
            )

            self.plot_lens_fit_pixelization_residual_map = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_pixelization_residual_map", bool
            )

            self.plot_lens_fit_pixelization_normalized_residuals = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_pixelization_normalized_residual_map", bool
            )

            self.plot_lens_fit_pixelization_chi_squared_map = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_pixelization_chi_squared_map", bool
            )

            self.plot_lens_fit_pixelization_regularization_weights = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_pixelization_regularization_weight_map", bool
            )

            self.plot_lens_fit_subtracted_images_of_planes = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_subtracted_images_of_planes", bool
            )

            self.plot_lens_fit_model_images_of_planes = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_model_images_of_planes", bool
            )

            self.plot_lens_fit_plane_images_of_planes = af.conf.instance.visualize.get(
                "plots", "plot_lens_fit_plane_images_of_planes", bool
            )

            self.use_inversion_border = use_inversion_border
            self.inversion_pixel_limit = inversion_pixel_limit

            mask = self.lens_data.mask_2d if self.should_plot_mask else None
            positions = self.lens_data.positions if self.should_plot_positions else None

            subplot_path = af.path_util.make_and_return_path_from_path_and_folder_names(
                path=image_path, folder_names=["subplots"]
            )

            phase_plotters.plot_ccd_for_phase(
                ccd_data=self.lens_data.ccd_data,
                mask=mask,
                positions=positions,
                extract_array_from_mask=self.extract_array_from_mask,
                zoom_around_mask=self.zoom_around_mask,
                units=self.plot_units,
                should_plot_as_subplot=self.plot_data_as_subplot,
                should_plot_image=self.plot_data_image,
                should_plot_noise_map=self.plot_data_noise_map,
                should_plot_psf=self.plot_data_psf,
                should_plot_signal_to_noise_map=self.plot_data_signal_to_noise_map,
                should_plot_absolute_signal_to_noise_map=self.plot_data_absolute_signal_to_noise_map,
                should_plot_potential_chi_squared_map=self.plot_data_potential_chi_squared_map,
                visualize_path=image_path,
                subplot_path=subplot_path,
            )

            self.plot_hyper_model_image = af.conf.instance.visualize.get(
                "plots", "plot_hyper_model_image", bool
            )

            self.plot_hyper_galaxy_images = af.conf.instance.visualize.get(
                "plots", "plot_hyper_galaxy_images", bool
            )

            self.plot_hyper_galaxy_cluster_images = af.conf.instance.visualize.get(
                "plots", "plot_hyper_galaxy_cluster_images", bool
            )

            self.pixelization_last = None

            if results is not None:
                if results.last is not None:
                    self.pixelization_last = results.last.most_likely_pixelization

            self.preload_pixelization_grid = None

            if self.last_results is not None:

                self.hyper_galaxy_image_1d_path_dict = (
                    self.last_results.hyper_galaxy_image_1d_path_dict
                )

                self.hyper_model_image_1d = self.last_results.hyper_model_image_1d

                self.hyper_galaxy_cluster_image_1d_path_dict = self.last_results.hyper_galaxy_cluster_image_1d_path_dict_from_cluster(
                    cluster=lens_data.cluster
                )

                if mask is not None:
                    phase_plotters.plot_hyper_images_for_phase(
                        hyper_model_image_2d=mask.scaled_array_2d_from_array_1d(
                            array_1d=self.hyper_model_image_1d
                        ),
                        hyper_galaxy_image_2d_path_dict=self.last_results.hyper_galaxy_image_2d_path_dict,
                        hyper_galaxy_cluster_image_2d_path_dict=self.last_results.hyper_galaxy_cluster_image_2d_path_dict_from_cluster(
                            cluster=lens_data.cluster
                        ),
                        mask=lens_data.mask_2d,
                        cluster=lens_data.cluster,
                        extract_array_from_mask=self.extract_array_from_mask,
                        zoom_around_mask=self.zoom_around_mask,
                        units=self.plot_units,
                        should_plot_hyper_model_image=self.plot_hyper_model_image,
                        should_plot_hyper_galaxy_images=self.plot_hyper_galaxy_images,
                        should_plot_hyper_galaxy_cluster_images=self.plot_hyper_galaxy_cluster_images,
                        visualize_path=image_path,
                    )

                if hasattr(self.results.last, "hyper_combined"):
                    self.preload_pixelization_grid = (
                        self.results.last.hyper_combined.most_likely_image_plane_pixelization_grid
                    )

        def fit(self, instance):
            """
            Determine the fit of a lens galaxy and source galaxy to the lens_data in this lens.

            Parameters
            ----------
            instance
                A model instance with attributes

            Returns
            -------
            fit : Fit
                A fractional value indicating how well this model fit and the model lens_data itself
            """
            self.check_positions_trace_within_threshold(instance=instance)
            self.check_inversion_pixels_are_below_limit(instance=instance)
            tracer = self.tracer_for_instance(instance=instance)

            hyper_image_sky = self.hyper_image_sky_for_instance(instance=instance)

            hyper_background_noise = self.hyper_background_noise_for_instance(
                instance=instance
            )

            fit = self.fit_for_tracer(
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            return fit.figure_of_merit

        def associate_images(self, instance: af.ModelInstance) -> af.ModelInstance:
            """
            Takes images from the last result, if there is one, and associates them with galaxies in this phase where
            full-path galaxy names match.

            If the galaxy collection has a different name then an association is not made.

            e.g.
            galaxies.lens will match with:
                galaxies.lens
            but not with:
                galaxies.lens
                galaxies.source

            Parameters
            ----------
            instance
                A model instance with 0 or more galaxies in its tree

            Returns
            -------
            instance
               The input instance with images associated with galaxies where possible.
            """
            if hasattr(self, "hyper_galaxy_image_1d_path_dict"):
                for galaxy_path, galaxy in instance.path_instance_tuples_for_class(
                    g.Galaxy
                ):
                    if galaxy_path in self.hyper_galaxy_image_1d_path_dict:
                        galaxy.hyper_model_image_1d = self.hyper_model_image_1d
                        galaxy.hyper_galaxy_image_1d = self.hyper_galaxy_image_1d_path_dict[
                            galaxy_path
                        ]
                        if (
                            hasattr(self, "hyper_galaxy_cluster_image_1d_path_dict")
                            and self.hyper_galaxy_cluster_image_1d_path_dict is not None
                        ):
                            galaxy.hyper_galaxy_cluster_image_1d = self.hyper_galaxy_cluster_image_1d_path_dict[
                                galaxy_path
                            ]
            return instance

        def add_grids_to_grid_stack(self, galaxies, grid_stack):

            pixelization = self.pixelization_for_galaxies(galaxies=galaxies)

            same_pixelizations = False

            if pixelization is not None and self.pixelization_last is not None:
                if type(pixelization) is type(self.pixelization_last):
                    same_pixelizations = True

            if self.preload_pixelization_grid is not None and same_pixelizations:

                return grid_stack.new_grid_stack_with_grids_added(
                    pixelization=self.preload_pixelization_grid
                )

            else:

                for galaxy in galaxies:
                    if galaxy.pixelization is not None:
                        pixelization_grid = galaxy.pixelization.pixelization_grid_from_grid_stack(
                            grid_stack=grid_stack,
                            hyper_image=galaxy.hyper_galaxy_cluster_image_1d,
                            cluster=self.lens_data.cluster,
                            seed=1,
                        )

                        return grid_stack.new_grid_stack_with_grids_added(
                            pixelization=pixelization_grid
                        )

            return grid_stack

        def hyper_image_sky_for_instance(self, instance):

            if hasattr(instance, "hyper_image_sky"):
                return instance.hyper_image_sky
            else:
                return None

        def hyper_background_noise_for_instance(self, instance):

            if hasattr(instance, "hyper_background_noise"):
                return instance.hyper_background_noise
            else:
                return None

        def tracer_for_instance(self, instance):

            instance = self.associate_images(instance=instance)

            image_plane_grid_stack = self.add_grids_to_grid_stack(
                galaxies=instance.galaxies, grid_stack=self.lens_data.grid_stack
            )

            if self.use_inversion_border:
                border = self.lens_data.border
            else:
                border = None

            return ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=instance.galaxies,
                image_plane_grid_stack=image_plane_grid_stack,
                border=border,
                cosmology=self.cosmology,
            )

        def fit_for_tracer(self, tracer, hyper_image_sky, hyper_background_noise):

            return lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=self.lens_data,
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

        def check_positions_trace_within_threshold(self, instance):

            if self.lens_data.positions is not None:

                tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_positions(
                    galaxies=instance.galaxies,
                    image_plane_positions=self.lens_data.positions,
                )
                fit = lens_fit.LensPositionFit(
                    positions=tracer.source_plane.positions,
                    noise_map=self.lens_data.pixel_scale,
                )

                if not fit.maximum_separation_within_threshold(
                    self.positions_threshold
                ):
                    raise exc.RayTracingException

        def check_inversion_pixels_are_below_limit(self, instance):

            if self.inversion_pixel_limit is not None:
                if instance.galaxies:
                    for galaxy in instance.galaxies:
                        if galaxy.pixelization is not None:
                            if galaxy.pixelization.pixels > self.inversion_pixel_limit:
                                raise exc.PixelizationException

        def map_to_1d(self, data):
            """Convenience method"""
            return self.lens_data.mask.array_1d_from_array_2d(data)

        @classmethod
        def describe(cls, instance):
            return "\nRunning for... \n\nGalaxies:\n{}\n\n".format(instance.galaxies)

        def visualize(self, instance, image_path, during_analysis):

            subplot_path = af.path_util.make_and_return_path_from_path_and_folder_names(
                path=image_path, folder_names=["subplots"]
            )

            instance = self.associate_images(instance=instance)

            mask = self.lens_data.mask_2d if self.should_plot_mask else None
            positions = self.lens_data.positions if self.should_plot_positions else None

            tracer = self.tracer_for_instance(instance)

            phase_plotters.plot_ray_tracing_for_phase(
                tracer=tracer,
                during_analysis=during_analysis,
                mask=mask,
                extract_array_from_mask=self.extract_array_from_mask,
                zoom_around_mask=self.zoom_around_mask,
                positions=positions,
                units=self.plot_units,
                should_plot_as_subplot=self.plot_ray_tracing_as_subplot,
                should_plot_all_at_end_png=self.plot_ray_tracing_all_at_end_png,
                should_plot_all_at_end_fits=self.plot_ray_tracing_all_at_end_fits,
                should_plot_image_plane_image=self.plot_ray_tracing_image_plane_image,
                should_plot_source_plane=self.plot_ray_tracing_source_plane,
                should_plot_convergence=self.plot_ray_tracing_convergence,
                should_plot_potential=self.plot_ray_tracing_potential,
                should_plot_deflections=self.plot_ray_tracing_deflections,
                visualize_path=image_path,
                subplot_path=subplot_path,
            )

            hyper_image_sky = self.hyper_image_sky_for_instance(instance=instance)

            hyper_background_noise = self.hyper_background_noise_for_instance(
                instance=instance
            )

            fit = self.fit_for_tracer(
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            phase_plotters.plot_lens_fit_for_phase(
                fit=fit,
                during_analysis=during_analysis,
                should_plot_mask=self.should_plot_mask,
                extract_array_from_mask=self.extract_array_from_mask,
                zoom_around_mask=self.zoom_around_mask,
                positions=positions,
                should_plot_image_plane_pix=self.should_plot_image_plane_pix,
                should_plot_all_at_end_png=self.plot_lens_fit_all_at_end_png,
                should_plot_all_at_end_fits=self.plot_lens_fit_all_at_end_fits,
                should_plot_fit_as_subplot=self.plot_lens_fit_as_subplot,
                should_plot_fit_of_planes_as_subplot=self.plot_lens_fit_of_planes_as_subplot,
                should_plot_inversion_as_subplot=self.plot_lens_fit_inversion_as_subplot,
                should_plot_image=self.plot_lens_fit_image,
                should_plot_noise_map=self.plot_lens_fit_noise_map,
                should_plot_signal_to_noise_map=self.plot_lens_fit_signal_to_noise_map,
                should_plot_model_image=self.plot_lens_fit_model_image,
                should_plot_residual_map=self.plot_lens_fit_residual_map,
                should_plot_normalized_residual_map=self.plot_lens_fit_normalized_residual_map,
                should_plot_chi_squared_map=self.plot_lens_fit_chi_squared_map,
                should_plot_pixelization_residual_map=self.plot_lens_fit_pixelization_residual_map,
                should_plot_pixelization_normalized_residual_map=self.plot_lens_fit_normalized_residual_map,
                should_plot_pixelization_chi_squared_map=self.plot_lens_fit_pixelization_chi_squared_map,
                should_plot_pixelization_regularization_weights=self.plot_lens_fit_pixelization_regularization_weights,
                should_plot_subtracted_images_of_planes=self.plot_lens_fit_subtracted_images_of_planes,
                should_plot_model_images_of_planes=self.plot_lens_fit_model_images_of_planes,
                should_plot_plane_images_of_planes=self.plot_lens_fit_plane_images_of_planes,
                units=self.plot_units,
                visualize_path=image_path,
                subplot_path=subplot_path,
            )

        def pixelization_for_galaxies(self, galaxies):
            for galaxy in galaxies:
                if galaxy.pixelization is not None:
                    return galaxy.pixelization



import autofit as af


class ScaledArrayException(Exception):
    pass


class GridException(Exception):
    pass


class DataException(Exception):
    pass


class KernelException(Exception):
    pass


class MaskException(Exception):
    pass


class CosmologyException(Exception):
    pass


class GalaxyException(Exception):
    pass


class RayTracingException(af.exc.FitException):
    pass


class PixelizationException(af.exc.FitException):
    pass


class InversionException(af.exc.FitException):
    pass


class FittingException(Exception):
    pass


class PlottingException(Exception):
    pass


class PhaseException(Exception):
    pass


class UnitsException(Exception):
    pass



from autofit import conf
from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.data.array.util import array_util
from autolens.plotters import array_plotters

import os
import numpy as np

slacs = []

slacs.append(
    {
        "name": "slacs0216-0813",
        "target_name": "GAL-0668-52162-428",
        "exposure_time": 2232.0,
        "redshift_lens": 0.332,
        "redshift_source": 0.523,
    }
)

path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))
conf.instance = af.conf.Config(config_path=path + "config", output_path=path + "output")

for slac in slacs:

    name = slac["name"]
    target_name = slac["target_name"]
    exposure_time = slac["exposure_time"]

    for filter in ["F555W"]:

        path = "{}/".format(os.path.dirname(os.path.realpath(__file__)))

        image_path = path + "/david_0/" + target_name + "_" + filter + "_post.fits"

        if os.path.exists(image_path):

            psf_path = path + "/david_0/" + target_name + "_" + filter + "_post.fits"
            background_noise_map_path = (
                path + "/david_0/" + target_name + "_" + filter + "_post.fits"
            )

            ccd_data = ccd.load_ccd_data_from_fits(
                image_path=image_path,
                image_hdu=1,
                pixel_scale=0.03,
                psf_path=psf_path,
                psf_hdu=3,
                noise_map_from_image_and_background_noise_map=True,
                background_noise_map_path=background_noise_map_path,
                background_noise_map_hdu=2,
                convert_background_noise_map_from_inverse_noise_map=True,
                poisson_noise_map_from_image=True,
                exposure_time_map_from_single_value=exposure_time,
                exposure_time_map_from_inverse_noise_map=True,
                resized_psf_shape=(61, 61),
            )

            ccd_data.noise_map = array_util.replace_noise_map_2d_values_where_image_2d_values_are_negative(
                image_2d=ccd_data.image, noise_map_2d=ccd_data.noise_map
            )

            print()
            print(name + "  " + filter + "  " + str(ccd_data.signal_to_noise_max))
            array_plotters.plot_array(array=1.0 / ccd_data.noise_map)
            # print(np.max(ccd_data.exposure_time_map))
            # print(np.max(ccd_data.poisson_noise_map))
            # print(np.max(ccd_data.noise_map))

            max_noise = np.max(ccd_data.noise_map)
            print()

            absolute_signal_to_noise = np.abs(ccd_data.image) / ccd_data.noise_map
            array_plotters.plot_array(absolute_signal_to_noise)
            print(np.max(absolute_signal_to_noise))

            ccd_data.noise_map[absolute_signal_to_noise > 200.0] = max_noise

            absolute_signal_to_noise = np.abs(ccd_data.image) / ccd_data.noise_map
            array_plotters.plot_array(absolute_signal_to_noise)
            stop
            #            ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

            path = path + "../../data/slacs/" + name + "/"

            if not os.path.exists(path):
                os.mkdir(path)

            ccd.output_ccd_data_to_fits(
                ccd_data=ccd_data,
                image_path=path + filter + "_image.fits",
                psf_path=path + filter + "_psf.fits",
                noise_map_path=path + filter + "_noise_map.fits",
                background_noise_map_path=path + filter + "_background_noise_map.fits",
                poisson_noise_map_path=path + filter + "_poisson_noise_map.fits",
                exposure_time_map_path=path + filter + "_exposure_time_map.fits",
                overwrite=True,
            )



import numpy as np
import pytest

import autofit as af
from autolens.data.array import mask as msk, scaled_array as sca
from autolens.model.galaxy import galaxy as g, galaxy_fit
from autolens.model.galaxy import galaxy_data as gd
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from test.unit.mock.model.mock_galaxy import MockGalaxy


class TestGalaxyFit:
    class TestLikelihood:
        def test__1x1_image__light_profile_fits_data_perfectly__lh_is_noise(self):
            image = sca.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)

            noise_map = sca.ScaledSquarePixelArray(
                array=np.ones((3, 3)), pixel_scale=1.0
            )

            galaxy_data = gd.GalaxyData(
                image=image, noise_map=noise_map, pixel_scale=3.0
            )

            mask = msk.Mask(
                array=np.array(
                    [[True, True, True], [True, False, True], [True, True, True]]
                ),
                pixel_scale=1.0,
            )
            g0 = MockGalaxy(value=1.0)

            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=galaxy_data,
                mask=mask,
                sub_grid_size=1,
                use_intensities=True,
            )
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=galaxy_data,
                mask=mask,
                sub_grid_size=1,
                use_convergence=True,
            )
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=galaxy_data, mask=mask, sub_grid_size=1, use_potential=True
            )
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=galaxy_data,
                mask=mask,
                sub_grid_size=1,
                use_deflections_y=True,
            )
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=galaxy_data,
                mask=mask,
                sub_grid_size=1,
                use_deflections_x=True,
            )
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

        def test__1x2_image__noise_not_1__alls_correct(self):
            image = sca.ScaledSquarePixelArray(
                array=5.0 * np.ones((3, 4)), pixel_scale=1.0
            )
            image[1, 2] = 4.0

            noise_map = sca.ScaledSquarePixelArray(
                array=2.0 * np.ones((3, 4)), pixel_scale=1.0
            )

            galaxy_data = gd.GalaxyData(
                image=image, noise_map=noise_map, pixel_scale=3.0
            )

            mask = msk.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            g0 = MockGalaxy(value=1.0, shape=2)

            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=galaxy_data,
                mask=mask,
                sub_grid_size=1,
                use_intensities=True,
            )
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])

            assert fit.model_galaxies == [g0]
            assert fit.chi_squared == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * (
                (25.0 / 4.0) + 2.0 * np.log(2 * np.pi * 2.0 ** 2)
            )

            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=galaxy_data,
                mask=mask,
                sub_grid_size=1,
                use_convergence=True,
            )
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.chi_squared == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * (
                (25.0 / 4.0) + 2.0 * np.log(2 * np.pi * 2.0 ** 2)
            )

            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=galaxy_data, mask=mask, sub_grid_size=1, use_potential=True
            )
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.model_galaxies == [g0]
            assert fit.chi_squared == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * (
                (25.0 / 4.0) + 2.0 * np.log(2 * np.pi * 2.0 ** 2)
            )

            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=galaxy_data,
                mask=mask,
                sub_grid_size=1,
                use_deflections_y=True,
            )
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.chi_squared == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * (
                (25.0 / 4.0) + 2.0 * np.log(2 * np.pi * 2.0 ** 2)
            )

            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=galaxy_data,
                mask=mask,
                sub_grid_size=1,
                use_deflections_x=True,
            )
            fit = galaxy_fit.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
            assert fit.chi_squared == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * (
                (25.0 / 4.0) + 2.0 * np.log(2 * np.pi * 2.0 ** 2)
            )

    class TestCompareToManual:
        def test__intensities(self, gal_data_7x7, mask_7x7):
            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=gal_data_7x7,
                mask=mask_7x7,
                sub_grid_size=2,
                use_intensities=True,
            )

            galaxy = g.Galaxy(
                redshift=0.5, light=lp.SphericalSersic(centre=(1.0, 2.0), intensity=1.0)
            )
            fit = galaxy_fit.GalaxyFit(
                galaxy_data=galaxy_fit_data, model_galaxies=[galaxy]
            )

            assert fit.model_galaxies == [galaxy]

            model_data_1d = galaxy.intensities_from_grid(
                grid=galaxy_fit_data.grid_stack.sub
            )
            model_data_1d = galaxy_fit_data.grid_stack.sub.array_1d_binned_from_sub_array_1d(
                sub_array_1d=model_data_1d
            )
            model_data_2d = galaxy_fit_data.map_to_scaled_array(array_1d=model_data_1d)
            residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=galaxy_fit_data.image(return_in_2d=True),
                mask=galaxy_fit_data.mask_2d,
                model_data=model_data_2d,
            )

            assert residual_map_2d == pytest.approx(
                fit.residual_map(return_in_2d=True), 1e-4
            )

            chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_2d,
                mask=galaxy_fit_data.mask_2d,
                noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
            )

            assert chi_squared_map_2d == pytest.approx(
                fit.chi_squared_map(return_in_2d=True), 1e-4
            )

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=mask_7x7
            )

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=galaxy_fit_data.mask_2d,
                noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
            )

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

        def test__convergence(self, gal_data_7x7, mask_7x7):
            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=gal_data_7x7,
                mask=mask_7x7,
                sub_grid_size=2,
                use_convergence=True,
            )

            galaxy = g.Galaxy(
                redshift=0.5,
                mass=mp.SphericalIsothermal(centre=(1.0, 2.0), einstein_radius=1.0),
            )
            fit = galaxy_fit.GalaxyFit(
                galaxy_data=galaxy_fit_data, model_galaxies=[galaxy]
            )

            assert fit.model_galaxies == [galaxy]

            model_data_1d = galaxy.convergence_from_grid(
                grid=galaxy_fit_data.grid_stack.sub
            )
            model_data_1d = galaxy_fit_data.grid_stack.sub.array_1d_binned_from_sub_array_1d(
                sub_array_1d=model_data_1d
            )
            model_data_2d = galaxy_fit_data.map_to_scaled_array(array_1d=model_data_1d)

            residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=galaxy_fit_data.image(return_in_2d=True),
                mask=galaxy_fit_data.mask_2d,
                model_data=model_data_2d,
            )
            assert residual_map_2d == pytest.approx(
                fit.residual_map(return_in_2d=True), 1e-4
            )

            chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_2d,
                mask=galaxy_fit_data.mask_2d,
                noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
            )
            assert chi_squared_map_2d == pytest.approx(
                fit.chi_squared_map(return_in_2d=True), 1e-4
            )

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=mask_7x7
            )

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=galaxy_fit_data.mask_2d,
                noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
            )

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

        def test__potential(self, gal_data_7x7, mask_7x7):
            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=gal_data_7x7,
                mask=mask_7x7,
                sub_grid_size=2,
                use_potential=True,
            )

            galaxy = g.Galaxy(
                redshift=0.5,
                mass=mp.SphericalIsothermal(centre=(1.0, 2.0), einstein_radius=1.0),
            )

            fit = galaxy_fit.GalaxyFit(
                galaxy_data=galaxy_fit_data, model_galaxies=[galaxy]
            )

            assert fit.model_galaxies == [galaxy]

            model_data_1d = galaxy.potential_from_grid(
                grid=galaxy_fit_data.grid_stack.sub
            )
            model_data_1d = galaxy_fit_data.grid_stack.sub.array_1d_binned_from_sub_array_1d(
                sub_array_1d=model_data_1d
            )
            model_data_2d = galaxy_fit_data.map_to_scaled_array(array_1d=model_data_1d)

            residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=galaxy_fit_data.image(return_in_2d=True),
                mask=galaxy_fit_data.mask_2d,
                model_data=model_data_2d,
            )

            assert residual_map_2d == pytest.approx(
                fit.residual_map(return_in_2d=True), 1e-4
            )

            chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_2d,
                mask=galaxy_fit_data.mask_2d,
                noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
            )

            assert chi_squared_map_2d == pytest.approx(
                fit.chi_squared_map(return_in_2d=True), 1e-4
            )

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=mask_7x7
            )

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=galaxy_fit_data.mask_2d,
                noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
            )

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

        def test__deflections_y(self, gal_data_7x7, mask_7x7):

            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=gal_data_7x7,
                mask=mask_7x7,
                sub_grid_size=2,
                use_deflections_y=True,
            )

            galaxy = g.Galaxy(
                redshift=0.5,
                mass=mp.SphericalIsothermal(centre=(1.0, 2.0), einstein_radius=1.0),
            )

            fit = galaxy_fit.GalaxyFit(
                galaxy_data=galaxy_fit_data, model_galaxies=[galaxy]
            )

            assert fit.model_galaxies == [galaxy]

            model_data_2d = galaxy.deflections_from_grid(
                grid=galaxy_fit_data.grid_stack.sub,
                return_in_2d=True,
                return_binned=True,
            )[:, :, 0]

            residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=galaxy_fit_data.image(return_in_2d=True),
                mask=galaxy_fit_data.mask_2d,
                model_data=model_data_2d,
            )

            assert residual_map_2d == pytest.approx(
                fit.residual_map(return_in_2d=True), 1e-4
            )

            chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_2d,
                mask=galaxy_fit_data.mask_2d,
                noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
            )

            assert chi_squared_map_2d == pytest.approx(
                fit.chi_squared_map(return_in_2d=True), 1e-4
            )

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=mask_7x7
            )

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=galaxy_fit_data.mask_2d,
                noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
            )

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

        def test__deflections_x(self, gal_data_7x7, mask_7x7):

            galaxy_fit_data = gd.GalaxyFitData(
                galaxy_data=gal_data_7x7,
                mask=mask_7x7,
                sub_grid_size=2,
                use_deflections_x=True,
            )

            galaxy = g.Galaxy(
                redshift=0.5,
                mass=mp.SphericalIsothermal(centre=(1.0, 2.0), einstein_radius=1.0),
            )
            fit = galaxy_fit.GalaxyFit(
                galaxy_data=galaxy_fit_data, model_galaxies=[galaxy]
            )

            assert fit.model_galaxies == [galaxy]

            model_data_2d = galaxy.deflections_from_grid(
                grid=galaxy_fit_data.grid_stack.sub,
                return_in_2d=True,
                return_binned=True,
            )[:, :, 1]

            residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=galaxy_fit_data.image(return_in_2d=True),
                mask=galaxy_fit_data.mask_2d,
                model_data=model_data_2d,
            )

            assert residual_map_2d == pytest.approx(
                fit.residual_map(return_in_2d=True), 1e-4
            )

            chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_2d,
                mask=galaxy_fit_data.mask_2d,
                noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
            )

            assert chi_squared_map_2d == pytest.approx(
                fit.chi_squared_map(return_in_2d=True), 1e-4
            )

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=mask_7x7
            )

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=galaxy_fit_data.mask_2d,
                noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
            )

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)



import os

import numpy as np
import pytest

from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.data.array import grids, mask as msk, scaled_array
from autolens.model.galaxy import galaxy as g
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion.plotters import mapper_plotters
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp


@pytest.fixture(name="mapper_plotter_path")
def make_mapper_plotter_setup():
    return "{}/../../../test_files/plotting/mapper/".format(
        os.path.dirname(os.path.realpath(__file__))
    )


@pytest.fixture(name="image")
def make_image():
    image = scaled_array.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)
    noise_map = ccd.NoiseMap(array=2.0 * np.ones((3, 3)), pixel_scale=1.0)
    psf = abstract_data.PSF(array=3.0 * np.ones((3, 3)), pixel_scale=1.0)

    return ccd.CCDData(image=image, pixel_scale=1.0, noise_map=noise_map, psf=psf)


@pytest.fixture(name="mask")
def make_mask():
    return msk.Mask.circular(shape=((3, 3)), pixel_scale=0.1, radius_arcsec=0.1)


@pytest.fixture(name="galaxy_light")
def make_galaxy_light():
    return g.Galaxy(redshift=0.5, light=lp.EllipticalSersic(intensity=1.0))


@pytest.fixture(name="galaxy_mass")
def make_galaxy_mass():
    return g.Galaxy(redshift=0.5, mass=mp.SphericalIsothermal(einstein_radius=1.0))


@pytest.fixture(name="grid_stack")
def make_grid_stack():
    return grids.GridStack.from_shape_pixel_scale_and_sub_grid_size(
        shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
    )


@pytest.fixture(name="border")
def make_border(mask):
    return grids.GridBorder.from_mask(mask=mask)


@pytest.fixture(name="rectangular_pixelization")
def make_rectangular_pixelization():
    return pix.Rectangular(shape=(25, 25))


@pytest.fixture(name="rectangular_mapper")
def make_rectangular_mapper(rectangular_pixelization, grid_stack, border):
    return rectangular_pixelization.mapper_from_grid_stack_and_border(
        grid_stack=grid_stack, border=border
    )


def test__image_and_rectangular_mapper_is_output(
    image, rectangular_mapper, mapper_plotter_path, plot_patch
):
    mapper_plotters.plot_image_and_mapper(
        ccd_data=image,
        mapper=rectangular_mapper,
        should_plot_centres=True,
        should_plot_grid=True,
        image_pixels=[[0, 1, 2], [3]],
        source_pixels=[[1, 2], [0]],
        output_path=mapper_plotter_path,
        output_format="png",
    )
    assert mapper_plotter_path + "image_and_mapper.png" in plot_patch.paths


def test__rectangular_mapper_is_output(
    rectangular_mapper, mapper_plotter_path, plot_patch
):
    mapper_plotters.plot_mapper(
        mapper=rectangular_mapper,
        should_plot_centres=True,
        should_plot_grid=True,
        image_pixels=[[0, 1, 2], [3]],
        source_pixels=[[1, 2], [0]],
        output_path=mapper_plotter_path,
        output_filename="rectangular_mapper",
        output_format="png",
    )
    assert mapper_plotter_path + "rectangular_mapper.png" in plot_patch.paths



import autofit as af
import matplotlib

backend = af.conf.instance.visualize.get("figures", "backend", str)
matplotlib.use(backend)
from matplotlib import pyplot as plt

from autolens.plotters import plotter_util
from autolens.lens.plotters import lens_plotter_util
from autolens.lens.plotters import plane_plotters
from autolens.model.inversion.plotters import inversion_plotters


def plot_fit_subplot(
    fit,
    should_plot_mask=True,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    should_plot_image_plane_pix=False,
    plot_mass_profile_centres=True,
    units="arcsec",
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_filename="lens_fit",
    output_format="show",
):

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=6
    )

    mask = lens_plotter_util.get_mask(fit=fit, should_plot_mask=should_plot_mask)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    kpc_per_arcsec = fit.tracer.image_plane.kpc_per_arcsec

    image_plane_pix_grid = lens_plotter_util.get_image_plane_pix_grid(
        should_plot_image_plane_pix, fit
    )

    lens_plotter_util.plot_image(
        fit=fit,
        grid=image_plane_pix_grid,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        position_pointsize=position_pointsize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 2)

    lens_plotter_util.plot_signal_to_noise_map(
        fit=fit,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        position_pointsize=position_pointsize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 3)

    lens_plotter_util.plot_model_data(
        fit=fit,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 4)

    lens_plotter_util.plot_residual_map(
        fit=fit,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 5)

    lens_plotter_util.plot_normalized_residual_map(
        fit=fit,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 6)

    lens_plotter_util.plot_chi_squared_map(
        fit=fit,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    plt.close()


def plot_fit_subplot_of_planes(
    fit,
    should_plot_mask=True,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    should_plot_image_plane_pix=False,
    plot_mass_profile_centres=True,
    units="arcsec",
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_filename="lens_fit_plane",
    output_format="show",
):

    for plane_index in range(fit.tracer.total_planes):

        if (
            fit.tracer.planes[plane_index].has_light_profile
            or fit.tracer.planes[plane_index].has_pixelization
        ):

            plot_fit_subplot_for_plane(
                fit=fit,
                plane_index=plane_index,
                should_plot_mask=should_plot_mask,
                extract_array_from_mask=extract_array_from_mask,
                zoom_around_mask=zoom_around_mask,
                should_plot_image_plane_pix=should_plot_image_plane_pix,
                positions=positions,
                units=units,
                figsize=figsize,
                aspect=aspect,
                cmap=cmap,
                norm=norm,
                norm_min=norm_min,
                norm_max=norm_max,
                linthresh=linthresh,
                linscale=linscale,
                cb_ticksize=cb_ticksize,
                cb_fraction=cb_fraction,
                cb_pad=cb_pad,
                cb_tick_values=cb_tick_values,
                cb_tick_labels=cb_tick_labels,
                titlesize=titlesize,
                xlabelsize=xlabelsize,
                ylabelsize=ylabelsize,
                xyticksize=xyticksize,
                grid_pointsize=grid_pointsize,
                position_pointsize=position_pointsize,
                mask_pointsize=mask_pointsize,
                output_path=output_path,
                output_filename=output_filename,
                output_format=output_format,
            )


def plot_fit_subplot_for_plane(
    fit,
    plane_index,
    should_plot_mask=True,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    should_plot_source_grid=False,
    positions=None,
    should_plot_image_plane_pix=False,
    plot_mass_profile_centres=True,
    units="arcsec",
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_filename="lens_fit_plane",
    output_format="show",
):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    output_filename += "_" + str(plane_index)

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=4
    )

    mask = lens_plotter_util.get_mask(fit=fit, should_plot_mask=should_plot_mask)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    kpc_per_arcsec = fit.tracer.image_plane.kpc_per_arcsec

    image_plane_pix_grid = lens_plotter_util.get_image_plane_pix_grid(
        should_plot_image_plane_pix, fit
    )

    plt.subplot(rows, columns, 1)

    lens_plotter_util.plot_image(
        fit=fit,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        grid=image_plane_pix_grid,
        positions=positions,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        position_pointsize=position_pointsize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 2)

    lens_plotter_util.plot_subtracted_image_of_plane(
        fit=fit,
        plane_index=plane_index,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        image_plane_pix_grid=image_plane_pix_grid,
        positions=positions,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 3)

    lens_plotter_util.plot_model_image_of_plane(
        fit=fit,
        plane_index=plane_index,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        plot_mass_profile_centres=plot_mass_profile_centres,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    if not fit.tracer.planes[plane_index].has_pixelization:

        plt.subplot(rows, columns, 4)

        plane_plotters.plot_plane_image(
            plane=fit.tracer.planes[plane_index],
            positions=None,
            plot_grid=should_plot_source_grid,
            as_subplot=True,
            units=units,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            grid_pointsize=grid_pointsize,
            position_pointsize=position_pointsize,
            output_path=output_path,
            output_filename="",
            output_format=output_format,
        )

    elif fit.tracer.planes[plane_index].has_pixelization:

        ratio = float(
            (
                fit.inversion.mapper.geometry.arc_second_maxima[1]
                - fit.inversion.mapper.geometry.arc_second_minima[1]
            )
            / (
                fit.inversion.mapper.geometry.arc_second_maxima[0]
                - fit.inversion.mapper.geometry.arc_second_minima[0]
            )
        )

        if aspect is "square":
            aspect_inv = ratio
        elif aspect is "auto":
            aspect_inv = 1.0 / ratio
        elif aspect is "equal":
            aspect_inv = 1.0

        plt.subplot(rows, columns, 4, aspect=float(aspect_inv))

        inversion_plotters.plot_pixelization_values(
            inversion=fit.inversion,
            positions=None,
            should_plot_grid=False,
            should_plot_centres=False,
            as_subplot=True,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            figsize=figsize,
            aspect=None,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            output_path=output_path,
            output_filename=None,
            output_format=output_format,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    plt.close()


def plot_fit_individuals(
    fit,
    should_plot_mask=True,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    should_plot_image_plane_pix=False,
    should_plot_image=False,
    should_plot_noise_map=False,
    should_plot_signal_to_noise_map=False,
    should_plot_model_image=False,
    should_plot_residual_map=False,
    should_plot_normalized_residual_map=False,
    should_plot_chi_squared_map=False,
    should_plot_pixelization_residual_map=False,
    should_plot_pixelization_normalized_residual_map=False,
    should_plot_pixelization_chi_squared_map=False,
    should_plot_pixelization_regularization_weight_map=False,
    should_plot_subtracted_images_of_planes=False,
    should_plot_model_images_of_planes=False,
    should_plot_plane_images_of_planes=False,
    units="arcsec",
    output_path=None,
    output_format="show",
):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    mask = lens_plotter_util.get_mask(fit=fit, should_plot_mask=should_plot_mask)
    image_plane_pix_grid = lens_plotter_util.get_image_plane_pix_grid(
        should_plot_image_plane_pix, fit
    )

    kpc_per_arcsec = fit.tracer.image_plane.kpc_per_arcsec

    if should_plot_image:

        lens_plotter_util.plot_image(
            fit=fit,
            mask=mask,
            positions=positions,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_noise_map:

        lens_plotter_util.plot_noise_map(
            fit=fit,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_signal_to_noise_map:

        lens_plotter_util.plot_signal_to_noise_map(
            fit=fit,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_model_image:

        lens_plotter_util.plot_model_data(
            fit=fit,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_residual_map:

        lens_plotter_util.plot_residual_map(
            fit=fit,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_normalized_residual_map:

        lens_plotter_util.plot_normalized_residual_map(
            fit=fit,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_chi_squared_map:

        lens_plotter_util.plot_chi_squared_map(
            fit=fit,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_pixelization_residual_map:

        if fit.total_inversions == 1:

            inversion_plotters.plot_pixelization_residual_map(
                inversion=fit.inversion,
                should_plot_grid=True,
                units=units,
                figsize=(20, 20),
                output_path=output_path,
                output_format=output_format,
            )

    if should_plot_pixelization_normalized_residual_map:

        if fit.total_inversions == 1:

            inversion_plotters.plot_pixelization_normalized_residual_map(
                inversion=fit.inversion,
                should_plot_grid=True,
                units=units,
                figsize=(20, 20),
                output_path=output_path,
                output_format=output_format,
            )

    if should_plot_pixelization_chi_squared_map:

        if fit.total_inversions == 1:

            inversion_plotters.plot_pixelization_chi_squared_map(
                inversion=fit.inversion,
                should_plot_grid=True,
                units=units,
                figsize=(20, 20),
                output_path=output_path,
                output_format=output_format,
            )

    if should_plot_pixelization_regularization_weight_map:

        if fit.total_inversions == 1:

            inversion_plotters.plot_pixelization_regularization_weights(
                inversion=fit.inversion,
                should_plot_grid=True,
                units=units,
                figsize=(20, 20),
                output_path=output_path,
                output_format=output_format,
            )

    if should_plot_subtracted_images_of_planes:

        for plane_index in range(fit.tracer.total_planes):

            lens_plotter_util.plot_subtracted_image_of_plane(
                fit=fit,
                plane_index=plane_index,
                mask=mask,
                extract_array_from_mask=extract_array_from_mask,
                zoom_around_mask=zoom_around_mask,
                units=units,
                kpc_per_arcsec=kpc_per_arcsec,
                output_path=output_path,
                output_format=output_format,
            )

    if should_plot_model_images_of_planes:

        for plane_index in range(fit.tracer.total_planes):

            lens_plotter_util.plot_model_image_of_plane(
                fit=fit,
                plane_index=plane_index,
                mask=mask,
                extract_array_from_mask=extract_array_from_mask,
                zoom_around_mask=zoom_around_mask,
                units=units,
                kpc_per_arcsec=kpc_per_arcsec,
                output_path=output_path,
                output_format=output_format,
            )

    if should_plot_plane_images_of_planes:

        for plane_index in range(fit.tracer.total_planes):

            if fit.tracer.planes[plane_index].has_light_profile:

                output_filename = "fit_plane_image_of_plane_" + str(plane_index)

                plane_plotters.plot_plane_image(
                    plane=fit.tracer.planes[plane_index],
                    plot_grid=True,
                    units=units,
                    output_path=output_path,
                    output_filename=output_filename,
                    output_format=output_format,
                )
